@@ -15,6 +15,7 @@ Other type classes are defined that help compose the types listed above.
 
 from __future__ import unicode_literals, division, absolute_import, print_function
 
+from contextlib import contextmanager
 from encodings import idna  # noqa
 import hashlib
 import re
@@ -443,14 +444,47 @@ class KeyUsage(BitString):
 
 class PrivateKeyUsagePeriod(Sequence):
     _fields = [
-        ('not_before', GeneralizedTime, {'tag_type': 'implicit', 'tag': 0, 'optional': True}),
-        ('not_after', GeneralizedTime, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
+        ('not_before', GeneralizedTime, {'implicit': 0, 'optional': True}),
+        ('not_after', GeneralizedTime, {'implicit': 1, 'optional': True}),
     ]
 
 
+class NotReallyTeletexString(TeletexString):
+    """
+    OpenSSL (and probably some other libraries) puts ISO-8859-1
+    into TeletexString instead of ITU T.61. We use Windows-1252 when
+    decoding since it is a superset of ISO-8859-1, and less likely to
+    cause encoding issues, but we stay strict with encoding to prevent
+    us from creating bad data.
+    """
+
+    _decoding_encoding = 'cp1252'
+
+    def __unicode__(self):
+        """
+        :return:
+            A unicode string
+        """
+
+        if self.contents is None:
+            return ''
+        if self._unicode is None:
+            self._unicode = self._merge_chunks().decode(self._decoding_encoding)
+        return self._unicode
+
+
+@contextmanager
+def strict_teletex():
+    try:
+        NotReallyTeletexString._decoding_encoding = 'teletex'
+        yield
+    finally:
+        NotReallyTeletexString._decoding_encoding = 'cp1252'
+
+
 class DirectoryString(Choice):
     _alternatives = [
-        ('teletex_string', TeletexString),
+        ('teletex_string', NotReallyTeletexString),
         ('printable_string', PrintableString),
         ('universal_string', UniversalString),
         ('utf8_string', UTF8String),
@@ -920,7 +954,7 @@ class Name(Choice):
 
         :param use_printable:
             A bool - if PrintableString should be used for encoding instead of
-            UTF8String. This is for backwards compatiblity with old software.
+            UTF8String. This is for backwards compatibility with old software.
 
         :return:
             An x509.Name object
@@ -1096,7 +1130,7 @@ class Name(Choice):
 class AnotherName(Sequence):
     _fields = [
         ('type_id', ObjectIdentifier),
-        ('value', Any, {'tag_type': 'explicit', 'tag': 0}),
+        ('value', Any, {'explicit': 0}),
     ]
 
 
@@ -1129,19 +1163,19 @@ class PrivateDomainName(Choice):
 
 class PersonalName(Set):
     _fields = [
-        ('surname', PrintableString, {'tag_type': 'implicit', 'tag': 0}),
-        ('given_name', PrintableString, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
-        ('initials', PrintableString, {'tag_type': 'implicit', 'tag': 2, 'optional': True}),
-        ('generation_qualifier', PrintableString, {'tag_type': 'implicit', 'tag': 3, 'optional': True}),
+        ('surname', PrintableString, {'implicit': 0}),
+        ('given_name', PrintableString, {'implicit': 1, 'optional': True}),
+        ('initials', PrintableString, {'implicit': 2, 'optional': True}),
+        ('generation_qualifier', PrintableString, {'implicit': 3, 'optional': True}),
     ]
 
 
 class TeletexPersonalName(Set):
     _fields = [
-        ('surname', TeletexString, {'tag_type': 'implicit', 'tag': 0}),
-        ('given_name', TeletexString, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
-        ('initials', TeletexString, {'tag_type': 'implicit', 'tag': 2, 'optional': True}),
-        ('generation_qualifier', TeletexString, {'tag_type': 'implicit', 'tag': 3, 'optional': True}),
+        ('surname', TeletexString, {'implicit': 0}),
+        ('given_name', TeletexString, {'implicit': 1, 'optional': True}),
+        ('initials', TeletexString, {'implicit': 2, 'optional': True}),
+        ('generation_qualifier', TeletexString, {'implicit': 3, 'optional': True}),
     ]
 
 
@@ -1157,13 +1191,13 @@ class BuiltInStandardAttributes(Sequence):
     _fields = [
         ('country_name', CountryName, {'optional': True}),
         ('administration_domain_name', AdministrationDomainName, {'optional': True}),
-        ('network_address', NumericString, {'tag_type': 'implicit', 'tag': 0, 'optional': True}),
-        ('terminal_identifier', PrintableString, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
-        ('private_domain_name', PrivateDomainName, {'tag_type': 'explicit', 'tag': 2, 'optional': True}),
-        ('organization_name', PrintableString, {'tag_type': 'implicit', 'tag': 3, 'optional': True}),
-        ('numeric_user_identifier', NumericString, {'tag_type': 'implicit', 'tag': 4, 'optional': True}),
-        ('personal_name', PersonalName, {'tag_type': 'implicit', 'tag': 5, 'optional': True}),
-        ('organizational_unit_names', OrganizationalUnitNames, {'tag_type': 'implicit', 'tag': 6, 'optional': True}),
+        ('network_address', NumericString, {'implicit': 0, 'optional': True}),
+        ('terminal_identifier', PrintableString, {'implicit': 1, 'optional': True}),
+        ('private_domain_name', PrivateDomainName, {'explicit': 2, 'optional': True}),
+        ('organization_name', PrintableString, {'implicit': 3, 'optional': True}),
+        ('numeric_user_identifier', NumericString, {'implicit': 4, 'optional': True}),
+        ('personal_name', PersonalName, {'implicit': 5, 'optional': True}),
+        ('organizational_unit_names', OrganizationalUnitNames, {'implicit': 6, 'optional': True}),
     ]
 
 
@@ -1223,8 +1257,8 @@ class UnformattedPostalAddress(Set):
 
 class E1634Address(Sequence):
     _fields = [
-        ('number', NumericString, {'tag_type': 'implicit', 'tag': 0}),
-        ('sub_address', NumericString, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
+        ('number', NumericString, {'implicit': 0}),
+        ('sub_address', NumericString, {'implicit': 1, 'optional': True}),
     ]
 
 
@@ -1234,17 +1268,17 @@ class NAddresses(SetOf):
 
 class PresentationAddress(Sequence):
     _fields = [
-        ('p_selector', OctetString, {'tag_type': 'explicit', 'tag': 0, 'optional': True}),
-        ('s_selector', OctetString, {'tag_type': 'explicit', 'tag': 1, 'optional': True}),
-        ('t_selector', OctetString, {'tag_type': 'explicit', 'tag': 2, 'optional': True}),
-        ('n_addresses', NAddresses, {'tag_type': 'explicit', 'tag': 3}),
+        ('p_selector', OctetString, {'explicit': 0, 'optional': True}),
+        ('s_selector', OctetString, {'explicit': 1, 'optional': True}),
+        ('t_selector', OctetString, {'explicit': 2, 'optional': True}),
+        ('n_addresses', NAddresses, {'explicit': 3}),
     ]
 
 
 class ExtendedNetworkAddress(Choice):
     _alternatives = [
         ('e163_4_address', E1634Address),
-        ('psap_address', PresentationAddress, {'tag_type': 'implicit', 'tag': 0})
+        ('psap_address', PresentationAddress, {'implicit': 0})
     ]
 
 
@@ -1289,8 +1323,8 @@ class ExtensionAttributeType(Integer):
 
 class ExtensionAttribute(Sequence):
     _fields = [
-        ('extension_attribute_type', ExtensionAttributeType, {'tag_type': 'implicit', 'tag': 0}),
-        ('extension_attribute_value', Any, {'tag_type': 'explicit', 'tag': 1}),
+        ('extension_attribute_type', ExtensionAttributeType, {'implicit': 0}),
+        ('extension_attribute_value', Any, {'explicit': 1}),
     ]
 
     _oid_pair = ('extension_attribute_type', 'extension_attribute_value')
@@ -1335,22 +1369,22 @@ class ORAddress(Sequence):
 
 class EDIPartyName(Sequence):
     _fields = [
-        ('name_assigner', DirectoryString, {'tag_type': 'implicit', 'tag': 0, 'optional': True}),
-        ('party_name', DirectoryString, {'tag_type': 'implicit', 'tag': 1}),
+        ('name_assigner', DirectoryString, {'implicit': 0, 'optional': True}),
+        ('party_name', DirectoryString, {'implicit': 1}),
     ]
 
 
 class GeneralName(Choice):
     _alternatives = [
-        ('other_name', AnotherName, {'tag_type': 'implicit', 'tag': 0}),
-        ('rfc822_name', EmailAddress, {'tag_type': 'implicit', 'tag': 1}),
-        ('dns_name', DNSName, {'tag_type': 'implicit', 'tag': 2}),
-        ('x400_address', ORAddress, {'tag_type': 'implicit', 'tag': 3}),
-        ('directory_name', Name, {'tag_type': 'explicit', 'tag': 4}),
-        ('edi_party_name', EDIPartyName, {'tag_type': 'implicit', 'tag': 5}),
-        ('uniform_resource_identifier', URI, {'tag_type': 'implicit', 'tag': 6}),
-        ('ip_address', IPAddress, {'tag_type': 'implicit', 'tag': 7}),
-        ('registered_id', ObjectIdentifier, {'tag_type': 'implicit', 'tag': 8}),
+        ('other_name', AnotherName, {'implicit': 0}),
+        ('rfc822_name', EmailAddress, {'implicit': 1}),
+        ('dns_name', DNSName, {'implicit': 2}),
+        ('x400_address', ORAddress, {'implicit': 3}),
+        ('directory_name', Name, {'explicit': 4}),
+        ('edi_party_name', EDIPartyName, {'implicit': 5}),
+        ('uniform_resource_identifier', URI, {'implicit': 6}),
+        ('ip_address', IPAddress, {'implicit': 7}),
+        ('registered_id', ObjectIdentifier, {'implicit': 8}),
     ]
 
     def __ne__(self, other):
@@ -1417,16 +1451,16 @@ class BasicConstraints(Sequence):
 
 class AuthorityKeyIdentifier(Sequence):
     _fields = [
-        ('key_identifier', OctetString, {'tag_type': 'implicit', 'tag': 0, 'optional': True}),
-        ('authority_cert_issuer', GeneralNames, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
-        ('authority_cert_serial_number', Integer, {'tag_type': 'implicit', 'tag': 2, 'optional': True}),
+        ('key_identifier', OctetString, {'implicit': 0, 'optional': True}),
+        ('authority_cert_issuer', GeneralNames, {'implicit': 1, 'optional': True}),
+        ('authority_cert_serial_number', Integer, {'implicit': 2, 'optional': True}),
     ]
 
 
 class DistributionPointName(Choice):
     _alternatives = [
-        ('full_name', GeneralNames, {'tag_type': 'implicit', 'tag': 0}),
-        ('name_relative_to_crl_issuer', RelativeDistinguishedName, {'tag_type': 'implicit', 'tag': 1}),
+        ('full_name', GeneralNames, {'implicit': 0}),
+        ('name_relative_to_crl_issuer', RelativeDistinguishedName, {'implicit': 1}),
     ]
 
 
@@ -1447,8 +1481,8 @@ class ReasonFlags(BitString):
 class GeneralSubtree(Sequence):
     _fields = [
         ('base', GeneralName),
-        ('minimum', Integer, {'tag_type': 'implicit', 'tag': 0, 'default': 0}),
-        ('maximum', Integer, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
+        ('minimum', Integer, {'implicit': 0, 'default': 0}),
+        ('maximum', Integer, {'implicit': 1, 'optional': True}),
     ]
 
 
@@ -1458,16 +1492,16 @@ class GeneralSubtrees(SequenceOf):
 
 class NameConstraints(Sequence):
     _fields = [
-        ('permitted_subtrees', GeneralSubtrees, {'tag_type': 'implicit', 'tag': 0, 'optional': True}),
-        ('excluded_subtrees', GeneralSubtrees, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
+        ('permitted_subtrees', GeneralSubtrees, {'implicit': 0, 'optional': True}),
+        ('excluded_subtrees', GeneralSubtrees, {'implicit': 1, 'optional': True}),
     ]
 
 
 class DistributionPoint(Sequence):
     _fields = [
-        ('distribution_point', DistributionPointName, {'tag_type': 'explicit', 'tag': 0, 'optional': True}),
-        ('reasons', ReasonFlags, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
-        ('crl_issuer', GeneralNames, {'tag_type': 'implicit', 'tag': 2, 'optional': True}),
+        ('distribution_point', DistributionPointName, {'explicit': 0, 'optional': True}),
+        ('reasons', ReasonFlags, {'implicit': 1, 'optional': True}),
+        ('crl_issuer', GeneralNames, {'implicit': 2, 'optional': True}),
     ]
 
     _url = False
@@ -1493,7 +1527,7 @@ class DistributionPoint(Sequence):
             for general_name in name.chosen:
                 if general_name.name == 'uniform_resource_identifier':
                     url = general_name.native
-                    if url[0:7] == 'http://':
+                    if url.lower().startswith(('http://', 'https://', 'ldap://', 'ldaps://')):
                         self._url = url
                         break
 
@@ -1585,8 +1619,8 @@ class PolicyMappings(SequenceOf):
 
 class PolicyConstraints(Sequence):
     _fields = [
-        ('require_explicit_policy', Integer, {'tag_type': 'implicit', 'tag': 0, 'optional': True}),
-        ('inhibit_policy_mapping', Integer, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
+        ('require_explicit_policy', Integer, {'implicit': 0, 'optional': True}),
+        ('inhibit_policy_mapping', Integer, {'implicit': 1, 'optional': True}),
     ]
 
 
@@ -1766,6 +1800,8 @@ class ExtensionId(ObjectIdentifier):
         '1.3.6.1.5.5.7.48.1.5': 'ocsp_no_check',
         '1.2.840.113533.7.65.0': 'entrust_version_extension',
         '2.16.840.1.113730.1.1': 'netscape_certificate_type',
+        # https://tools.ietf.org/html/rfc6962.html#page-14
+        '1.3.6.1.4.1.11129.2.4.2': 'signed_certificate_timestamp_list',
     }
 
 
@@ -1800,6 +1836,7 @@ class Extension(Sequence):
         'ocsp_no_check': Null,
         'entrust_version_extension': EntrustVersionInfo,
         'netscape_certificate_type': NetscapeCertificateType,
+        'signed_certificate_timestamp_list': OctetString,
     }
 
 
@@ -1817,16 +1854,16 @@ class Version(Integer):
 
 class TbsCertificate(Sequence):
     _fields = [
-        ('version', Version, {'tag_type': 'explicit', 'tag': 0, 'default': 'v1'}),
+        ('version', Version, {'explicit': 0, 'default': 'v1'}),
         ('serial_number', Integer),
         ('signature', SignedDigestAlgorithm),
         ('issuer', Name),
         ('validity', Validity),
         ('subject', Name),
         ('subject_public_key_info', PublicKeyInfo),
-        ('issuer_unique_id', OctetBitString, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
-        ('subject_unique_id', OctetBitString, {'tag_type': 'implicit', 'tag': 2, 'optional': True}),
-        ('extensions', Extensions, {'tag_type': 'explicit', 'tag': 3, 'optional': True}),
+        ('issuer_unique_id', OctetBitString, {'implicit': 1, 'optional': True}),
+        ('subject_unique_id', OctetBitString, {'implicit': 2, 'optional': True}),
+        ('extensions', Extensions, {'explicit': 3, 'optional': True}),
     ]
 
 
@@ -2374,7 +2411,7 @@ class Certificate(Sequence):
                 if location.name != 'uniform_resource_identifier':
                     continue
                 url = location.native
-                if url.lower()[0:7] == 'http://':
+                if url.lower().startswith(('http://', 'https://', 'ldap://', 'ldaps://')):
                     output.append(url)
         return output
 
@@ -2518,6 +2555,16 @@ class Certificate(Sequence):
             self._sha256 = hashlib.sha256(self.dump()).digest()
         return self._sha256
 
+    @property
+    def sha256_fingerprint(self):
+        """
+        :return:
+            A unicode string of the SHA-256 hash, formatted using hex encoding
+            with a space between each pair of characters, all uppercase
+        """
+
+        return ' '.join('%02X' % c for c in bytes_to_list(self.sha256))
+
     def is_valid_domain_ip(self, domain_ip):
         """
         Check if a domain name or IP address is valid according to the
@@ -2670,10 +2717,10 @@ class SequenceOfAlgorithmIdentifiers(SequenceOf):
 class CertificateAux(Sequence):
     _fields = [
         ('trust', KeyPurposeIdentifiers, {'optional': True}),
-        ('reject', KeyPurposeIdentifiers, {'tag_type': 'implicit', 'tag': 0, 'optional': True}),
+        ('reject', KeyPurposeIdentifiers, {'implicit': 0, 'optional': True}),
         ('alias', UTF8String, {'optional': True}),
         ('keyid', OctetString, {'optional': True}),
-        ('other', SequenceOfAlgorithmIdentifiers, {'tag_type': 'implicit', 'tag': 1, 'optional': True}),
+        ('other', SequenceOfAlgorithmIdentifiers, {'implicit': 1, 'optional': True}),
     ]
 
 
